/*
 ** xkaapi
 ** 
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 ** thierry.gautier@inrialpes.fr
 ** fabien.lementec@imag.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
#include "kaapi_impl.h"
#include "kaapic_impl.h"

/*
*/
int kaapic_foreach_attr_init(kaapic_foreach_attr_t* attr)
{
  attr->init  = 0;
  return 0;
}

/*
*/
int kaapic_foreach_attr_set_policy(
  kaapic_foreach_attr_t* attr, 
  kaapic_foreach_attr_policy_t policy
)
{
  kaapi_assert_debug( (policy == KAAPIC_FOREACH_SCHED_STEAL)
                    ||(policy == KAAPIC_FOREACH_SCHED_DISTRIBUTE))
  if( (policy != KAAPIC_FOREACH_SCHED_STEAL)
    &&(policy != KAAPIC_FOREACH_SCHED_DISTRIBUTE))
    return EINVAL;
  attr->policy = policy;
  attr->init |= KAAPIC_FOREACH_ATTR_INIT_POLICY;
  return 0;
}

/*
*/
int kaapic_foreach_attr_get_policy(
  const kaapic_foreach_attr_t* attr,
  kaapic_foreach_attr_policy_t* policy
)
{
  if (attr->init & KAAPIC_FOREACH_ATTR_INIT_POLICY)
    *policy = attr->policy;
  else
    *policy = KAAPIC_FOREACH_SCHED_DEFAULT;
  return 0;
}

/*
*/
int kaapic_foreach_attr_set_grains(
  kaapic_foreach_attr_t* attr, 
  long s_grain,
  long p_grain
)
{
  if ((s_grain <1) || (p_grain <1))
    return EINVAL;
  attr->rep.li.s_grain = s_grain;
  attr->rep.li.p_grain = p_grain;
  attr->init |= KAAPIC_FOREACH_ATTR_INIT_SGRAIN | KAAPIC_FOREACH_ATTR_INIT_PGRAIN;
  return 0;
}

/*
*/
int kaapic_foreach_attr_get_grains(
  const kaapic_foreach_attr_t* attr,
  long* s_grain,
  long* p_grain
)
{
  if (attr->init & KAAPIC_FOREACH_ATTR_INIT_SGRAIN)
    *s_grain = attr->rep.li.s_grain;
  else
    *s_grain = 1;

  if (attr->init & KAAPIC_FOREACH_ATTR_INIT_PGRAIN)
    *p_grain = attr->rep.li.p_grain;
  else
    *p_grain = 1;
  return 0;
}

/*
*/
int kaapic_foreach_attr_set_grains_ull(
  kaapic_foreach_attr_t* attr, 
  unsigned long long s_grain,
  unsigned long long p_grain
)
{
  attr->rep.ull.s_grain = s_grain;
  attr->rep.ull.p_grain = p_grain;
  attr->init |= KAAPIC_FOREACH_ATTR_INIT_SGRAIN | KAAPIC_FOREACH_ATTR_INIT_PGRAIN;
  return 0;
}

/*
*/
int kaapic_foreach_attr_get_grains_ull(
  const kaapic_foreach_attr_t* attr,
  unsigned long long* s_grain,
  unsigned long long* p_grain
)
{
  if (attr->init & KAAPIC_FOREACH_ATTR_INIT_SGRAIN)
    *s_grain = attr->rep.ull.s_grain;
  else
    *s_grain = 1;

  if (attr->init & KAAPIC_FOREACH_ATTR_INIT_PGRAIN)
    *p_grain = attr->rep.ull.p_grain;
  else
    *p_grain = 1;
  return 0;
}

/*
*/
int kaapic_foreach_attr_set_nthreads(
  kaapic_foreach_attr_t* attr, 
  unsigned int nthreads
)
{
  if (nthreads <1)
    return EINVAL;
  attr->nthreads = nthreads;
  attr->init |= KAAPIC_FOREACH_ATTR_INIT_NTHREADS;
  return 0;
}

/*
*/
int kaapic_foreach_attr_get_nthreads(
  const kaapic_foreach_attr_t* attr, 
  unsigned int* nthreads
)
{
  if (attr->init & KAAPIC_FOREACH_ATTR_INIT_NTHREADS)
    *nthreads = attr->nthreads;
  else
    *nthreads = kaapi_getconcurrency();
  return 0;
}

/*
*/
int kaapic_foreach_attr_set_cpuset(
  kaapic_foreach_attr_t* attr, 
  const kaapi_cpuset_t* threadset
)
{
  if (kaapi_cpuset_empty(threadset))
    return EINVAL;
  attr->threadset = *threadset;
  attr->init |= KAAPIC_FOREACH_ATTR_INIT_CPUSET;
  return 0;
}

/*
*/
int kaapic_foreach_attr_get_cpuset(
  kaapic_foreach_attr_t* attr, 
  kaapi_cpuset_t*        threadset
)
{
  if (attr->init & KAAPIC_FOREACH_ATTR_INIT_CPUSET)
    *threadset = attr->threadset;
  else
  {
    kaapi_cpuset_clear(threadset);
    kaapi_bitmap_value_set_low_bits((kaapi_bitmap_value_t*)&threadset, kaapi_getconcurrency());
  }
  return 0;
}


/*
*/
int kaapic_foreach_attr_set_bloccyclic_datadistribution(
  kaapic_foreach_attr_t* attr, 
  unsigned long long blocsize,
  unsigned int cyclelength
)
{
  return ENOSYS;
}
