/*
 ** xkaapi
 **
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 **
 ** thierry.gautier@inrialpes.fr
 ** Joao.Lima@imagf.r / joao.lima@inf.ufrgs.br
 **
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 **
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 **
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 **
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 **
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 **
 */

#include <stdio.h>
#include <pthread.h>
#include <cuda_runtime_api.h>

#include "kaapi_impl.h"
#include "kaapi_cuda_impl.h"

int kaapi_cuda_ctx_init(kaapi_cuda_ctx_t* const ctx, const unsigned int dev)
{
  cudaError_t res;
  
  //#if KAAPI_VERBOSE
#if 0
  fprintf(stdout, "%s: dev=%lu kid=%lu\n", __FUNCTION__, dev, kaapi_get_self_kid());
  fflush(stdout);
#endif
  ctx->dev = dev;
  if ((res = kaapi_cuda_dev_open(ctx)) != cudaSuccess)
    return res;
  
  kaapi_cuda_device_sync();
  
  if (kaapi_default_param.cudawindowsize > 0)
    kaapi_cuda_stream_init(kaapi_default_param.cudawindowsize * 3, ctx);
  else
    kaapi_cuda_stream_init(512, ctx);
  
  kaapi_cuda_cublas_init(ctx);
  kaapi_cuda_cublas_set_stream();
  
  if (kaapi_default_param.cudapeertopeer)
    kaapi_cuda_dev_enable_peer_access(ctx);
  
  kaapi_cuda_mem_cache_init(ctx);
  kaapi_cuda_device_sync();
  
#if defined(KAAPI_USE_CUPTI)
#warning JOAO need to review
  if (getenv("KAAPI_RECORD_TRACE") != 0) {
    kaapi_cuda_trace_thread_init();
  }
  
  kaapi_atomic_initlock(&ctx->lock);
#endif 
  
  return 0;
}

kaapi_cuda_ctx_t* kaapi_cuda_ctx_get_current(void)
{
  kaapi_processor_t* const kproc = kaapi_get_current_processor();
  const int index = kaapi_processor_get_cudaproc(kproc)->index;
  return &kaapi_processor_get_cudaproc(kproc)->ctx[index];
}

int kaapi_cuda_ctx_get_current_device(void)
{
  kaapi_cuda_ctx_t* const ctx = kaapi_cuda_ctx_get_current();
  return ctx->dev;
}

#if defined(KAAPI_USE_CUPTI)
void kaapi_cuda_ctx_set_(const int dev)
{
  kaapi_processor_t* const kproc = kaapi_cuda_get_proc_by_dev(dev);
  kaapi_atomic_lock(&kaapi_processor_get_cudaproc(kproc)->ctx.lock);
}

void kaapi_cuda_ctx_exit_(const int dev)
{
  kaapi_processor_t* const kproc = kaapi_cuda_get_proc_by_dev(dev);
  kaapi_atomic_unlock(&kaapi_processor_get_cudaproc(kproc)->ctx.lock);
}
#endif /* KAAPI_USE_CUPTI */

void kaapi_cuda_ctx_set(const int dev)
{
#if defined(KAAPI_USE_CUPTI)
  kaapi_cuda_ctx_set_(dev);
#endif
  const cudaError_t res = cudaSetDevice(dev);
  if (res != cudaSuccess) {
    fprintf(stderr, "%s: ERROR %d\n", __FUNCTION__, res);
    fflush(stderr);
  }
}

void kaapi_cuda_ctx_exit(const int dev)
{
#if defined(KAAPI_USE_CUPTI)
  kaapi_cuda_ctx_exit_(dev);
#endif
}

void kaapi_cuda_ctx_set_current_(kaapi_processor_t* const kproc, const int index)
{
  kaapi_processor_get_cudaproc(kproc)->index= index;
  kaapi_cuda_ctx_set(kaapi_processor_get_cudaproc(kproc)->ctx[index].dev);
}

