/*
** xkaapi
** 
** Created on Jan 2012
** Copyright 2010 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** Joao.Lima@imag.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/

#ifndef KAAPI_CUDA_TRACE_H_INCLUDED
#define KAAPI_CUDA_TRACE_H_INCLUDED

#if defined(KAAPI_USE_PERFCOUNTER) && defined(KAAPI_USE_CUPTI)

#include <cuda_runtime_api.h>
#include <cupti.h>

#include "kaapi_impl.h"
#include "kaapi_event.h"

/* Init globally CUDA traces */
int kaapi_cuda_trace_init(void);

/* Init traces by CUDA thread */
int kaapi_cuda_trace_thread_init(void);

/* Finalize globally CUDA traces */
void kaapi_cuda_trace_finalize(void);

/* Finalize traces by CUDA thread */
void kaapi_cuda_trace_thread_finalize(void);

#else

static inline int kaapi_cuda_trace_init(void)
{
  return 0;
}

static inline int kaapi_cuda_trace_thread_init(void)
{
  return 0;
}

static inline void kaapi_cuda_trace_finalize(void)
{
}

static inline void kaapi_cuda_trace_thread_finalize(void)
{
}

#endif


#endif				/* KAAPI_CUDA_TRACE_H_INCLUDED */
