/*
 ** xkaapi
 **
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 **
 ** thierry.gautier@inrialpes.fr
 ** Joao.Lima@imagf.r / joao.lima@inf.ufrgs.br
 **
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 **
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 **
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 **
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 **
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 **
 */

#include "kaapi_impl.h"

static inline kaapi_data_t* kaapi_memory_task_prologue_alloc(
                                                            kaapi_metadata_info_t* kmdi,
                                                            kaapi_address_space_id_t kasid,
                                                            kaapi_memory_view_t* view,
                                                            const kaapi_access_mode_t m
                                                            )
{
  kaapi_pointer_t ptr = kaapi_memory_allocate_view(kasid, view, (int)m);
#if 0
  fprintf(stdout, "%d:%s: kmdi=%p data=%p size=%d\n", kaapi_get_self_kid(), __FUNCTION__,
          (void*)kmdi, kaapi_pointer2void(ptr), (int)kaapi_memory_view_size(view));
  fflush(stdout);
#endif
  kaapi_memory_bind_view_with_metadata(kasid, kmdi, 0, kaapi_pointer2void(ptr), view);
  kaapi_metadata_info_set_dirty(kmdi, kasid);
  
  return kaapi_metadata_info_get_data(kmdi, kasid);
}

static inline kaapi_data_t* kaapi_memory_task_prologue_access(
                                                             kaapi_metadata_info_t* kmdi,
                                                             kaapi_address_space_id_t kasid,
                                                             const kaapi_access_mode_t m
                                                             )
{
  kaapi_data_t* kdest = kaapi_metadata_info_get_data(kmdi, kasid);
  kaapi_memory_increase_access_view(kasid, &kdest->ptr, &kdest->view, (int)m);
  
  return kdest;
}

static inline int kaapi_memory_task_prologue_transfer(kaapi_metadata_info_t* const kmdi, kaapi_data_t* kdest)
{
  if(kaapi_metadata_info_is_valid(kmdi, kaapi_pointer2asid(kdest->ptr)))
    return 0;
  
  uint16_t lid = kaapi_metadata_info_first_valid(kmdi);
  kaapi_data_t* ksrc = kaapi_metadata_info_get_data_by_lid(kmdi, lid);
  kaapi_memory_copy(kdest->ptr, &kdest->view, ksrc->ptr, &ksrc->view);
#if 1
  fprintf(stdout, "%d:%s: kmdi=@%p dest=@%p(%d) src=@%p(%d) size=%lu\n",
          kaapi_get_self_kid(), __FUNCTION__, (void*)kmdi,
          kaapi_pointer2void(kdest->ptr), kaapi_memory_address_space_getlid(kaapi_pointer2asid(kdest->ptr)),
          kaapi_pointer2void(ksrc->ptr), kaapi_memory_address_space_getlid(kaapi_pointer2asid(ksrc->ptr)),
          kaapi_memory_view_size(&kdest->view)
          );
  fflush(stdout);
#endif
  kaapi_metadata_info_clear_dirty(kmdi, kaapi_pointer2asid(kdest->ptr));
  
  return 0;
}

static inline void kaapi_memory_task_prologue_validate(
                                                        const kaapi_format_t* fmt,
                                                        void* sp,
                                                        const int i,
                                                        const kaapi_access_mode_t m
                                                        )
{
  kaapi_access_t access = kaapi_format_get_access_param(fmt, i, sp);
  kaapi_memory_view_t view = kaapi_format_get_view_param(fmt, i, sp);
  kaapi_data_t* kdata;
  kaapi_address_space_id_t kasid = kaapi_memory_map_get_current_asid();
  const kaapi_address_space_id_t host_kasid = KAAPI_EMPTY_ADDRESS_SPACE_ID;  
  
  kaapi_metadata_info_t* const kmdi = kaapi_memory_find_metadata_by_asid(host_kasid, access.data, kaapi_memory_view_size(&view));
//  kaapi_metadata_info_t* const kmdi = kaapi_memory_find_metadata(access.data);
  if(!kaapi_metadata_info_has_data(kmdi, kasid))
  {
    if(!kaapi_metadata_info_has_data(kmdi, host_kasid))
    {
#if 0
      fprintf(stdout, "%d:%s: BIND kmdi=%p data=%p size=%d\n", kaapi_get_self_kid(), __FUNCTION__,
              (void*)kmdi, access.data, kaapi_memory_view_size(&view));
      fflush(stdout);
#endif
      kaapi_memory_bind_view(host_kasid, 0, access.data, &view);
    }
    else
    {
#if 0
      fprintf(stdout, "%d:%s:  FOUND kmdi=%p data=%p size=%d\n", kaapi_get_self_kid(), __FUNCTION__,
              (void*)kmdi, access.data, kaapi_memory_view_size(&view));
      fflush(stdout);
#endif
    }
    
    if(!kaapi_memory_address_space_isequal(kasid, host_kasid))
    {
      kaapi_memory_task_prologue_alloc(kmdi, kasid, &view, m);
    }
  }
  
  kdata = kaapi_metadata_info_get_data(kmdi, kasid);
#if 0
  fprintf(stdout,
          "%d:%s: src=%p %ldx%ld lda=%ld size=%ld\n",
          kaapi_get_self_kid(),
          __FUNCTION__,
          __kaapi_pointer2void(kdata->ptr), kdata->view.size[0], kdata->view.size[1],
          kdata->view.lda, kaapi_memory_view_size(&kdata->view));
  fflush(stdout);
#endif
//  kaapi_memory_task_prologue_access(kasid, m);
  
  if (KAAPI_ACCESS_IS_READ(m))
  {
    kaapi_memory_task_prologue_transfer(kmdi, kdata);
  }
  
  if (KAAPI_ACCESS_IS_WRITE(m))
  {
    kaapi_metadata_info_set_all_dirty_except(kmdi, kasid);
  }
  
  if(kaapi_memory_address_space_gettype(kasid) != KAAPI_MEM_TYPE_CPU)
  {
    /* sets new pointer to the task */
    /* TODO correct here */
    access.data = kaapi_pointer2void(kdata->ptr);;
    kaapi_format_set_access_param(fmt, i, sp, &access);
    kaapi_format_set_view_param(fmt, i, sp, &kdata->view);
  }
}

int kaapi_memory_task_prologue(const kaapi_format_t* fmt, void* sp)
{
  size_t i;
  
  if (fmt == 0)
    return 0;
  
  const size_t count_params = kaapi_format_get_count_params(fmt, sp);
#if 0
  fprintf(stdout, "%d:%s: params=%ld asid=%d\n", kaapi_get_self_kid(), __FUNCTION__,  
          count_params,
          kaapi_memory_address_space_getlid(kaapi_memory_map_get_current_asid()));
  fflush(stdout);
#endif
  
  for (i = 0; i < count_params; i++) {
    kaapi_access_mode_t m = kaapi_format_get_mode_param(fmt, i, sp);
    m = KAAPI_ACCESS_GET_MODE(m);
    if (m & KAAPI_ACCESS_MODE_V)
      continue;
    
    kaapi_memory_task_prologue_validate(fmt, sp, i, m);
  }
  
  return 0;
}

static inline void kaapi_memory_task_epilogue_validate(
                                                      const kaapi_format_t* fmt,
                                                      void* sp,
                                                      const int i,
                                                      const kaapi_access_mode_t m
                                                      )
{
  kaapi_access_t access = kaapi_format_get_access_param(fmt, i, sp);
  kaapi_data_t* ksrc = kaapi_data(kaapi_data_t, &access);
  kaapi_address_space_id_t kasid = kaapi_memory_map_get_current_asid();
  
  kaapi_memory_decrease_access_view(kasid, &ksrc->ptr, &ksrc->view, (int)m);
}

int kaapi_memory_task_epilogue(const kaapi_format_t* fmt, void* sp)
{
  size_t i;
  
  if(fmt == 0)
    return 0;
  
  const size_t count_params = kaapi_format_get_count_params(fmt, sp);
#if 0
  fprintf(stdout, "%d:%s: params=%ld asid=%d\n", kaapi_get_self_kid(), __FUNCTION__,
          count_params,
          kaapi_memory_address_space_getlid(kaapi_memory_map_get_current_asid()));
  fflush(stdout);
#endif
  
  for (i = 0; i < count_params; i++) {
    kaapi_access_mode_t m = kaapi_format_get_mode_param(fmt, i, sp);
    
    m = KAAPI_ACCESS_GET_MODE(m);
    if (m & KAAPI_ACCESS_MODE_V)
      continue;
    
//    kaapi_memory_task_epilogue_validate(fmt, sp, i, m);
  }
  
  return 0;
}

