/*
 ** xkaapi
 **
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 **
 ** thierry.gautier@inrialpes.fr
 ** Joao.Lima@imagf.r / joao.lima@inf.ufrgs.br
 **
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 **
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 **
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 **
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 **
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 **
 */

#ifndef _KAAPI_MEMORY_TASKDESCR_H_
#define _KAAPI_MEMORY_TASKDESCR_H_

#include "kaapi_impl.h"
#include "kaapi_tasklist.h"

/* return the sum of invalid data on the kproc in bytes */
extern uint64_t kaapi_memory_taskdescr_affinity_get_invalid_bytes(
                                                                  struct kaapi_processor_t* const kproc,
                                                                  kaapi_taskdescr_t* const td
                                                                  );

extern uint64_t kaapi_memory_taskdescr_get_footprint_(kaapi_taskdescr_t* const td);

/* return the sum of task data in bytes */
static inline uint64_t kaapi_memory_taskdescr_get_footprint(kaapi_taskdescr_t* const td)
{
  if(td->footprint != 0)
    return td->footprint;
  
  return kaapi_memory_taskdescr_get_footprint_(td);
}

/* return true if td has an WR parameter valid in kproc
 */
extern int kaapi_memory_taskdescr_has_valid_writer( const struct kaapi_processor_t* kproc, kaapi_taskdescr_t * td );

/**
 Return a processor where a W/WR parameter of the task is valid.
 Otherwise, return kproc.
 */
extern struct kaapi_processor_t *kaapi_memory_taskdescr_affinity_find_valid_wr( struct kaapi_processor_t * kproc, kaapi_taskdescr_t * td );

/**
 Return a processor that reduces memory transfers to execute this task.
 Otherwise, return kproc.
 */
extern struct kaapi_processor_t *kaapi_memory_taskdescr_affinity_find_reduce_transfer( struct kaapi_processor_t * kproc, kaapi_taskdescr_t * td );

int kaapi_memory_taskdescr_prologue(kaapi_taskdescr_t * td);

int kaapi_memory_taskdescr_epilogue(kaapi_taskdescr_t * td);

#endif /* _KAAPI_MEMORY_TASKDESCR_H_ */
