/*
** xkaapi
** 
** Copyright 2013 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/

#include "kaapi_impl.h"

/* lock must have been taken 
*/
kaapi_task_t* kaapi_queue_steal( kaapi_queue_t* q, kaapi_processor_id_t kid )
{
  kaapi_taskbloc_t* bloc;
  int pos;

  KAAPI_DEBUG_INST(kaapi_atomic_assertlocked( &q->lock );)

  bloc = q->tail;
  while (bloc !=0)
  {  
    /* if empty bloc, then move it to freelist */
    if (bloc->top == bloc->bottom)
    {
      bloc->top = bloc->bottom = 0;
      if (bloc->prev ==0) q->head = 0;
      else bloc->prev->next = 0;
      q->tail = bloc->prev;

      bloc->next = q->freelist;
      q->freelist = bloc;
      goto next_iteration;
    }
    
    /* pop task until can mark it steal first */
    for (pos =bloc->bottom; pos< bloc->top; ++pos)
    {
      kaapi_task_t* task = bloc->data[pos];
      if ((task !=0) && kaapi_task_isstealable(task) && (!kaapi_task_is_strictaffinity(task) || kaapi_cpuset_has(&q->cpuset, kid)))
      {
        int retval = kaapi_task_marksteal( task );
        bloc->data[pos] = 0;
        if (pos == bloc->bottom) ++bloc->bottom;
        /* here only look at execution state, not signaled nor locked */
        if (likely( (retval & KAAPI_TASK_EXECUTION_STATE) ) ) 
        {
//printf("[queue::steal] id: %i, lid:%i, steal task: %p\n", q->id, q->lid, (void*)task );
          return task;
        }
      }
    }
next_iteration:
    bloc = bloc->prev;
  }
  return 0;
}
