
#ifndef _KAAPI_PERFMODEL_H_
#define _KAAPI_PERFMODEL_H_

#include "kaapi_impl.h"

/* Task and communication model for performance prediction (HEFT and GDUAL).
  All time results are in microseconds (us).
*/

extern void kaapi_perfmodel_task_init(void);

extern void kaapi_perfmodel_task_finalize(void);

static inline int kaapi_perfmodel_calibrate_is_enabled(void)
{
  return (kaapi_default_param.perfmodel_calibrate != 0);
}

/* Return the predicted execution time in microseconds (us) */
extern double kaapi_perfmodel_task_predict(kaapi_processor_t* const kproc,
                                           struct kaapi_taskdescr_t* const td);

/* Return the predicted execution time for architecture arch in microseconds (us) */
extern double kaapi_perfmodel_task_predict_by_arch(const int arch, struct kaapi_taskdescr_t* const td);


/* Call before task execution to measure execution time */
extern void kaapi_perfmodel_task_prologue(kaapi_processor_t* const kproc,
                              struct kaapi_taskdescr_t* const td);

/* Call after task execution to measure execution time */
extern void kaapi_perfmodel_task_epilogue(kaapi_processor_t* const kproc,
                                          struct kaapi_taskdescr_t* const td);

extern void kaapi_perfmodel_comm_init(void);

extern void kaapi_perfmodel_comm_finalize(void);

extern double kaapi_perfmodel_comm_predict_transfer(kaapi_processor_t* const kproc,
                                                    struct kaapi_taskdescr_t* const td);

extern void kaapi_perfmodel_init(void);

extern void kaapi_perfmodel_finalize(void);

extern void kaapi_perfmodel_print_stats(void);

static inline void kaapi_perfmodel_prologue(kaapi_processor_t* const kproc,
                                           struct kaapi_taskdescr_t* const td)
{
  kaapi_perfmodel_task_prologue(kproc, td);
}

static inline void kaapi_perfmodel_epilogue(kaapi_processor_t* const kproc,
                                           struct kaapi_taskdescr_t* const td)
{
  kaapi_perfmodel_task_epilogue(kproc, td);
}

/** Update a communication entry of the performance model.
 \param src_asid: address space source.
 \param dst_asid: address space destiny.
 \param size: data transfered in bytes.
 \param time: time spent on transfer.
 */
extern void kaapi_perfmodel_comm_update(const kaapi_address_space_id_t src_asid, const kaapi_address_space_id_t dst_asid,
                                                      const size_t size, const double time);


#endif /* _KAAPI_PERFMODEL_H_ */
