
#include "kaapi_impl.h"

/** Push one ready task to the correct queue.
 * 0) it tests the site is set
 * 1) it tests the task priority and
 * 2) affinity (if enabled)
 * 3) them pushes in the readylist rtl
 */
static inline int kaapi_sched_locality_writer_pushready_td(kaapi_readytasklist_t* const rtl, kaapi_taskdescr_t* const td)
{
  kaapi_processor_t* curr_kproc = kaapi_get_current_processor();
  
  if(td->fmt == 0)
  {
    return kaapi_readylist_push(rtl, td);
  }
  
  if (kaapi_taskdescr_get_site(td) !=0)
  {
    return kaapi_sched_push_by_site(td);    
  }
  
#if defined(KAAPI_USE_CUDA)
  /* TG: why to not do that for any kind of processor ?
   */
  if (kaapi_processor_get_type(curr_kproc) == KAAPI_PROC_TYPE_CUDA)
  {
    kaapi_processor_t* kproc_remote = kaapi_memory_taskdescr_affinity_find_valid_wr(curr_kproc, td);
    if ( kproc_remote != curr_kproc )
    {
      return kaapi_readylist_remote_push( kproc_remote->rtl, td );
    }
    
    if ( (kaapi_get_task_bodywh_by_arch(td, KAAPI_PROC_TYPE_CPU) == 0) ||
        (kaapi_task_has_arch(td->task, KAAPI_PROC_TYPE_CPU) == 0) )
    {
      return kaapi_readylist_push( curr_kproc->rtl, td );
    }
  }
#endif

  /* if td does not have GPU implementation, push it on master tasklist */
  if ( (kaapi_get_task_bodywh_by_arch(td, kaapi_processor_get_type(curr_kproc)) == 0) ||
      (kaapi_task_has_arch(td->task, kaapi_processor_get_type(curr_kproc)) == 0) )
  {
    kaapi_assert_debug( td->tasklist != 0 );
    return kaapi_readylist_push( &td->tasklist->rtl, td );
  }
  
  return kaapi_readylist_push( rtl, td );
}

static inline int kaapi_sched_locality_writer_push_from_activationlist(kaapi_readytasklist_t* const rtl, kaapi_activationlink_t*	head)
{
  kaapi_taskdescr_t* td;
  int retval =0;
  
  while (head !=0) {
    td = head->td;
    if (kaapi_taskdescr_activated(td)) {
      ++retval;
      kaapi_sched_locality_writer_pushready_td(rtl, td);
    }
    head = head->next;
  }
  return retval;
}

static inline uint32_t kaapi_sched_locality_writer_pushactivated(kaapi_readytasklist_t* const rtl, kaapi_taskdescr_t* const td)
{
  uint32_t cnt_pushed= 0;
  
  /* push in the front the activated tasks */
  if (!kaapi_activationlist_isempty(&td->u.acl.list))
    cnt_pushed = kaapi_sched_locality_writer_push_from_activationlist( rtl, td->u.acl.list.front );
  else
    cnt_pushed = 0;
  
  /* do bcast after child execution (they can produce output data) */
  if (td->u.acl.bcast !=0)
    cnt_pushed += kaapi_sched_locality_writer_push_from_activationlist( rtl, td->u.acl.bcast->front );
  
  return cnt_pushed;
}

void kaapi_sched_locality_writer_init(void)
{
  
}

void kaapi_sched_locality_writer_finalize(void)
{
  
}

uint32_t kaapi_sched_locality_writer_push_activated( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  return kaapi_sched_locality_writer_pushactivated( &tasklist->rtl, td );  
}

void kaapi_sched_locality_writer_push( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  kaapi_sched_locality_writer_pushready_td(&tasklist->rtl, td);
}

void kaapi_sched_locality_writer_task_epilogue(kaapi_taskdescr_t* const td)
{
  
}

void kaapi_sched_locality_writer_task_prologue(kaapi_taskdescr_t* const td)
{
  
}
