/*
** kaapi_sched_sync.c
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include "kaapi_impl.h"

#define NANO_SLEEP 0
#if NANO_SLEEP
#if !defined(_XOPEN_SOURCE)
#  define _XOPEN_SOURCE 600
#endif
#include <time.h>
#warning "Use nanosleep in sched_lock"
#endif

/**/
#define KAAPI_GET_THREAD_STATE(kproc)\
  ((kproc)->curr_perf_regs == (kproc)->perf_regs[KAAPI_PERF_USER_STATE] ? KAAPI_PERF_USER_STATE : KAAPI_PERF_SCHEDULE_STATE)

/**
*/
int kaapi_sched_initlock( kaapi_lock_t* lock )
{
  return kaapi_atomic_initlock(lock);
}

int kaapi_sched_destroylock( kaapi_lock_t* lock )
{
  return kaapi_atomic_destroylock(lock);
}

int kaapi_sched_trylock( kaapi_lock_t* lock )
{
  int retval;
  retval = kaapi_atomic_trylock(lock);  
  return retval;
}

/** 
*/
int kaapi_sched_lock( kaapi_lock_t* lock )
{
  int retval;
  
#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_perf_counter_t date0=0,date1=0;
  kaapi_processor_t* kproc=0;
  if (kaapi_default_param.display_perfcounter !=0)
  {
    kproc = kaapi_get_current_processor();
    date0 = kaapi_get_elapsedns();
  }
#endif

#if NANO_SLEEP
  kaapi_processor_t* kproc;
  kproc = kaapi_get_current_processor();
  struct timespec ts;
  ts.tv_sec  = 0;
  ts.tv_nsec = rand_r(&kproc->seed) % 50;
  while (!kaapi_atomic_trylock(lock))
  {
    clock_nanosleep(CLOCK_REALTIME, 0, &ts, 0);
    ts.tv_nsec = rand_r(&kproc->seed) % 50;
  }
  retval = 0;
#else
  retval = kaapi_atomic_lock(lock);
#endif

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kaapi_default_param.display_perfcounter !=0)
  {
    date1 = kaapi_get_elapsedns();
    KAAPI_PERF_REG(kproc, KAAPI_PERF_ID_TSYNC) += date1-date0;
  }
#endif
  return retval;
}

/**
*/
int kaapi_sched_unlock( kaapi_lock_t* lock )
{
  int retval;
  retval= kaapi_atomic_unlock(lock);
  return retval;
}

void kaapi_sched_waitlock( kaapi_lock_t* lock)
{
#if defined(KAAPI_USE_PERFCOUNTER)
  kaapi_perf_counter_t date0=0, date1=0;
  kaapi_processor_t* kproc=0;
  if (kaapi_default_param.display_perfcounter !=0)
  {
    kproc = kaapi_get_current_processor();
    date0 = kaapi_get_elapsedns();
  }
#endif

  kaapi_atomic_waitlock(lock);

#if defined(KAAPI_USE_PERFCOUNTER)
  if (kaapi_default_param.display_perfcounter !=0)
  {
    date1 = kaapi_get_elapsedns();
    KAAPI_PERF_REG(kproc, KAAPI_PERF_ID_TSYNC) += date1-date0;
  }
#endif
}
