/*
 ** xkaapi
 ** 
 ** Created on Tue Mar 31 15:21:00 2009
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributors :
 **
 ** thierry.gautier@inrialpes.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */

#include "kaapi_impl.h"
#include "machine/cuda/kaapi_cuda_impl.h"

typedef struct kaapi_hws_gpu_t {
  struct kaapi_hws_gpu_t* next;
  int dev;
} kaapi_hws_gpu_t;

static int kaapi_level1[4][2] = {
  {0, 1}, {2, 3}, {4, 5}, {6, 7}
};

static int kaapi_level2[2][4] = {
  {0, 1, 2, 3}, {4, 5, 6, 7}
};

static inline int kaapi_sched_select_victim_hws_gpu_from_level1(kaapi_processor_t* kproc)
{
  const unsigned int dev = kaapi_cuda_self_device();
  unsigned int devid;
  const int i = (int)(dev/2);
  const int j = rand_r( (unsigned int*)&kproc->fnc_selecarg ) % 2;
  devid = kaapi_level1[i][j];
  if(dev == devid)
  {
    const int jj = (j+1) % 2;
    devid = kaapi_level1[i][jj];
  }
#if 0
  fprintf(stdout, "%s: dev=%d i=%d j=%d victim=%d\n", __FUNCTION__, dev, i, j, devid );
  fflush(stdout);
#endif
  return devid;
}

static inline int kaapi_sched_select_victim_hws_gpu_from_level2(kaapi_processor_t* kproc)
{
  const unsigned int dev = kaapi_cuda_self_device();
  unsigned int devid;  
  const int i = (int)(dev/4);
  const int j = rand_r( (unsigned int*)&kproc->fnc_selecarg ) % 4;
  devid = kaapi_level2[i][j];
  if(dev == devid)
  {
    const int jj = (j+1) % 4;
    devid = kaapi_level2[i][jj];
  }
#if 0
  fprintf(stdout, "%s: dev=%d i=%d j=%d victim=%d\n", __FUNCTION__, dev, i, j, devid );
  fflush(stdout);
#endif
  return devid;
}

static inline int kaapi_sched_select_victim_hws_gpu_from_level(
                                                               kaapi_processor_t* kproc,
                                                               kaapi_victim_t* victim,
                                                               int level
                                                               )
{
  unsigned int devid;
  
  if(level == 0)
    devid = kaapi_sched_select_victim_hws_gpu_from_level1(kproc);
  else
    devid = kaapi_sched_select_victim_hws_gpu_from_level2(kproc);
  
  victim->kproc = kaapi_cuda_get_proc_by_dev(devid);
  if((victim->kproc == 0) && (level == 0))
  {
    devid = kaapi_sched_select_victim_hws_gpu_from_level2(kproc);
    victim->kproc = kaapi_cuda_get_proc_by_dev(devid);    
  }
  if (victim->kproc == 0)
    return EINVAL;
  
  return 0;
}

int kaapi_sched_select_victim_hws_gpu(
                                   kaapi_processor_t* kproc, 
                                   kaapi_victim_t* victim, 
                                   kaapi_selecvictim_flag_t flag 
                                   )
{
  if (kaapi_processor_get_type(kproc) == KAAPI_PROC_TYPE_CPU)
    return kaapi_sched_select_victim_rand(kproc, victim, flag);
  
  const int level = (int)kproc->fnc_selecarg[1];
  
  switch (flag)
  {
    case KAAPI_SELECT_VICTIM:
    {
      if(level > 1)
      {
        kaapi_sched_select_victim_rand(kproc, victim, flag);
      }
      else
      {
        if(kaapi_sched_select_victim_hws_gpu_from_level(kproc, victim, level) != 0)
        {
          kaapi_sched_select_victim_rand(kproc, victim, flag);          
        }
      }
      break;
    }
      
    case KAAPI_STEAL_FAILED:
    {
      kproc->fnc_selecarg[1] = (uintptr_t)(level+1);
      break;
    }
      
    case KAAPI_STEAL_SUCCESS:
    {
      kproc->fnc_selecarg[1] = (uintptr_t)0;
      break;
    }
      
    default:
    {
      break;
    }
  }
  
  return 0;
}
