/*
** xkaapi
** 
**
** Copyright 2009,2010,2011,2012 INRIA.
**
** Contributors :
**
** thierry.gautier@inrialpes.fr
** 
** This software is a computer program whose purpose is to execute
** multithreaded computation with data flow synchronization between
** threads.
** 
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software.  You can
** use, modify and/ or redistribute the software under the terms of
** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
** following URL "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to
** copy, modify and redistribute granted by the license, users are
** provided only with a limited warranty and the software's author,
** the holder of the economic rights, and the successive licensors
** have only limited liability.
** 
** In this respect, the user's attention is drawn to the risks
** associated with loading, using, modifying and/or developing or
** reproducing the software by the user in light of its specific
** status of free software, that may mean that it is complicated to
** manipulate, and that also therefore means that it is reserved for
** developers and experienced professionals having in-depth computer
** knowledge. Users are therefore encouraged to load and test the
** software's suitability as regards their requirements in conditions
** enabling the security of their systems and/or data to be ensured
** and, more generally, to use and operate it in the same conditions
** as regards security.
** 
** The fact that you are presently reading this means that you have
** had knowledge of the CeCILL-C license and that you accept its
** terms.
** 
*/
#include "kaapi_impl.h"
#include <stdlib.h>


/* cf macro name ! */
//#define JE_VEUX_VOLER_MES_COUSINS 1


/* This is an very experimental code. 
   Principle in this version:
   - the cores are splited into sets with respect to NUMA node definition
   - an idle core only steal in its NUMA node set
*/


typedef struct kaapi_hier_arg {
  int            init;
  int            nextkid;  /* next kid for the next selection */
  int            next;     /* next entry for the next selection */
  int            iter0;    /* current iteration number in NUMA level */
  int            iter1;    /* current iteration at NUMA level */
  unsigned int   seed;
} kaapi_hier_arg;

int kaapi_sched_select_victim_pws(
    kaapi_processor_t* kproc,
    kaapi_victim_t* victim,
    kaapi_selecvictim_flag_t flag
)
{
  int i, nbproc, victimid;
  int numalevel;
  kaapi_hier_arg* arg;


  numalevel = kproc->numalevel;
  nbproc = kproc->hlevel.levels[numalevel].nkids;
  if (nbproc <= 1) return EINVAL;

#if 1
  arg = (kaapi_hier_arg*)&kproc->fnc_selecarg;

  if (flag == KAAPI_STEAL_FAILED)
  {
#if defined(JE_VEUX_VOLER_MES_COUSINS)
    /* do it only after first success steal ! */
    if (arg->iter1 >0)
    {
      ++arg->iter1;
      //int rand_cousin = rand_r(&kproc->seed) % 1000000;
      if (((arg->iter1+1) % 1000) ==0)
      {
        int cousin;
        while ((numalevel < kproc->hlevel.depth) && (kproc->hlevel.levels[numalevel].nnotself ==0))
          ++numalevel;
        if (kproc->hlevel.levels[numalevel].nnotself !=0) 
        {
          cousin = rand_r(&kproc->seed) % kproc->hlevel.levels[numalevel].nnotself;
          victimid = kproc->hlevel.levels[numalevel].notself[ cousin ];
          arg->nextkid = victimid;
//printf("[pws] cousin case ! %i -> %i\n", kproc->kid, victimid );
        }
      }
    }
#endif
    return 0;
  } 
  if (flag == KAAPI_STEAL_SUCCESS)
  {
    arg->iter1 = 0;  
    if (arg->iter1 == 0)
      ++arg->iter1;  
    return 0;
  }

  if (arg->init ==0)
  {
    arg->init = 1;
    arg->next = rand_r(&kproc->seed) % nbproc;
    arg->nextkid = kproc->hlevel.levels[numalevel].kids[ arg->next ];
    arg->next = 0;
    arg->iter0 = 0;
    arg->iter1 = 0;
  }

  victim->kproc = 0;
  for (i=0; i<nbproc; ++i)
  {
    victimid  = arg->nextkid;

    if (++arg->next >= nbproc)
      arg->next = 0;
    if (++arg->iter0 >= nbproc)
    { /* new seed */ 
      arg->next    = rand_r(&kproc->seed) % nbproc;
      arg->iter0 = 0;
    }
    arg->nextkid = kproc->hlevel.levels[numalevel].kids[ arg->next ];

    victim->kproc = kaapi_all_kprocessors[ victimid ];
    if (victim->kproc != 0)
      return 0;
  }
  return EINVAL;
#else
  for (i=0; i<nbproc; ++i)
  {
    victimid = kproc->hlevel.levels[numalevel].kids[ rand_r( (unsigned int*)&kproc->seed ) % nbproc ];
    victim->kproc = kaapi_all_kprocessors[ victimid ];
    if (victim->kproc != 0)
      return 0;
  }
  return EINVAL;
#endif
}


