/*
 ** xkaapi
 ** 
 **
 ** Copyright 2009,2010,2011,2012 INRIA.
 **
 ** Contributor :
 **
 ** thierry.gautier@inrialpes.fr
 ** Joao.Lima@imag.fr
 ** 
 ** This software is a computer program whose purpose is to execute
 ** multithreaded computation with data flow synchronization between
 ** threads.
 ** 
 ** This software is governed by the CeCILL-C license under French law
 ** and abiding by the rules of distribution of free software.  You can
 ** use, modify and/ or redistribute the software under the terms of
 ** the CeCILL-C license as circulated by CEA, CNRS and INRIA at the
 ** following URL "http://www.cecill.info".
 ** 
 ** As a counterpart to the access to the source code and rights to
 ** copy, modify and redistribute granted by the license, users are
 ** provided only with a limited warranty and the software's author,
 ** the holder of the economic rights, and the successive licensors
 ** have only limited liability.
 ** 
 ** In this respect, the user's attention is drawn to the risks
 ** associated with loading, using, modifying and/or developing or
 ** reproducing the software by the user in light of its specific
 ** status of free software, that may mean that it is complicated to
 ** manipulate, and that also therefore means that it is reserved for
 ** developers and experienced professionals having in-depth computer
 ** knowledge. Users are therefore encouraged to load and test the
 ** software's suitability as regards their requirements in conditions
 ** enabling the security of their systems and/or data to be ensured
 ** and, more generally, to use and operate it in the same conditions
 ** as regards security.
 ** 
 ** The fact that you are presently reading this means that you have
 ** had knowledge of the CeCILL-C license and that you accept its
 ** terms.
 ** 
 */
#include "kaapi_impl.h"


/** Steal ready task in tasklist.
    Steal on rtl and rtl_remote that contains taskdescr_t.
    This function should disapear and must be merge with kaapi_sched_stealreadytasklist
    in order to have only one set of queue with task_t and not taskdescr_t with task_t.
 */
void kaapi_sched_stealreadytasklist( 
                           kaapi_thread_context_t*       thread, 
                           kaapi_processor_t*            kproc,
                           kaapi_listrequest_t*          lrequests, 
                           kaapi_listrequest_iterator_t* lrrange
)
{
  int                         err;
  kaapi_readytasklist_t*      rtl;
  kaapi_taskdescr_t*          td;
  kaapi_task_t*               tasksteal;
  kaapi_request_t*            request;
  kaapi_taskstealready_arg_t* argsteal;
    
  request = kaapi_listrequest_iterator_get( lrequests, lrrange );
  while( request !=0 )
  {
    rtl = kproc->rtl;
    err = kaapi_readylist_steal( kproc, rtl, &td );
    if( err == 0 )
    {
      /* - create the task steal that will execute the stolen task
        The task stealtask stores:
           - the original thread
           - the original task pointer
           - the pointer to shared data with R / RW access data
           - and at the end it reserve enough space to store original task arguments
        The original body is saved as the extra body of the original task data structure.
      */
      argsteal = (kaapi_taskstealready_arg_t*)kaapi_request_pushdata(request, sizeof(kaapi_taskstealready_arg_t));
      argsteal->master_tasklist = td->tasklist;
      argsteal->td              = td; /* recopy of the pointer, need synchro if range */

      tasksteal = kaapi_request_toptask(request);
      kaapi_task_init(  tasksteal, kaapi_taskstealready_body, argsteal );
      kaapi_request_pushtask(request, 0);

      kaapi_request_replytask( request, KAAPI_REQUEST_S_OK); /* success of steal */
      KAAPI_DEBUG_INST( kaapi_listrequest_iterator_countreply( lrrange ) );

      /* update next request to process */
      kaapi_listrequest_iterator_next( lrequests, lrrange );    
      request = kaapi_listrequest_iterator_get( lrequests, lrrange );
    }
    else
      break;
  }
}


void kaapi_sched_stealreadytasklist_priority(
                           kaapi_thread_context_t*       thread, 
                           kaapi_processor_t*            kproc,
                           kaapi_listrequest_t*          lrequests, 
                           kaapi_listrequest_iterator_t* lrrange
)
{
  int                         i;
  kaapi_task_t*               task;
  kaapi_task_t*               tasksteal;
  kaapi_request_t*            request;
  kaapi_tasksteal_arg_t*      argsteal;
    
  request = kaapi_listrequest_iterator_get( lrequests, lrrange );
  for (i=KAAPI_HWS_LEVELID_MACHINE; (i>= KAAPI_HWS_LEVELID_L0) && (request !=0); --i)
  {
    if (   (kproc->highpriority_queue[i] !=0)
        && !kaapi_queue_isempty(kproc->highpriority_queue[i])
        /* here because when core becomes idle, during the call to kaapi_sched_popreadytasklist_priority,
           the idle core has already tries to pop task in its hierarchy of queues...
        */
        && !kaapi_cpuset_has(&kproc->highpriority_queue[i]->cpuset, kproc->kid)
    )
    {
      kaapi_sched_lock( &kproc->highpriority_queue[i]->lock);
      task = kaapi_queue_steal( kproc->highpriority_queue[i], request->ident );
      kaapi_sched_unlock( &kproc->highpriority_queue[i]->lock);
      if (task !=0)
      {  
        /* push the task into local queue */
        /* queue_steal return a pointer to the task marked as steal. The task must be embeded into
           a kaapi_task_steal. Nevertheless, if the task is a task_steal, then reset the flag.
           This case occurs only (in this version X.Y.Z) when using kaapi_thread_distribute_task:
           the local task is pushed on the local queue of the running thread, while the remote
           task is push on the remote queue only. This task is never shared between multiple queue !
        */
        if (task->body != kaapi_tasksteal_body)
        {
          argsteal = (kaapi_tasksteal_arg_t*)kaapi_request_pushdata(request, sizeof(kaapi_tasksteal_arg_t));
          argsteal->origin_task           = task;
          argsteal->origin_body           = task->body;
          argsteal->origin_fmt            = 0;
          argsteal->war_param             = 0;    /* assume no war param: must be computed by thief ? */
          argsteal->cw_param              = 0;    /* assume no cw param: must be computed by thief ? */
          
          tasksteal = kaapi_request_toptask(request);
          kaapi_task_init( 
            tasksteal,
            kaapi_tasksteal_body,
            argsteal
          );
          ((kaapi_task_t* volatile)task)->reserved = tasksteal;
        }
        else {
          tasksteal = kaapi_request_toptask(request);
          kaapi_task_init( 
            tasksteal,
            task->body,
            task->sp
          );
          ((kaapi_task_t* volatile)task)->reserved = tasksteal;
        }

        kaapi_request_pushtask(request, 0);
        kaapi_request_replytask( request, KAAPI_REQUEST_S_OK); /* success of steal */
        KAAPI_DEBUG_INST( kaapi_listrequest_iterator_countreply( lrrange ) );
        /* update next request to process */
        kaapi_listrequest_iterator_next( lrequests, lrrange );    
        request = kaapi_listrequest_iterator_get( lrequests, lrrange );
        if (request ==0) break;
        continue;
      }
    }
  } // for
}



int kaapi_sched_popreadytasklist_priority(
                           kaapi_thread_context_t*       thctxt,
                           kaapi_processor_t*            kproc
)
{
  int                         i;
  kaapi_task_t*               task;
  kaapi_task_t*               tasksteal;
  kaapi_thread_t*             thread;
  kaapi_tasksteal_arg_t*      argsteal;
  
  thread = kaapi_threadcontext2thread( thctxt );
  
  for (i=KAAPI_HWS_LEVELID_L0; i< KAAPI_HWS_LEVELID_MAX; ++i)
  {
    if ((kproc->highpriority_queue[i] !=0) && !kaapi_queue_isempty(kproc->highpriority_queue[i]) )
    {
      kaapi_sched_lock( &kproc->highpriority_queue[i]->lock);
      task = kaapi_queue_pop( kproc->highpriority_queue[i] );
      kaapi_sched_unlock( &kproc->highpriority_queue[i]->lock);
      if (task !=0)
      {  
        /* push the task into local queue */
        /* queue_steal return a pointer to the task marked as steal. The task must be embeded into
           a kaapi_task_steal. Nevertheless, if the task is a task_steal, then reset the flag.
           This case occurs only (in this version X.Y.Z) when using kaapi_thread_distribute_task:
           the local task is pushed on the local queue of the running thread, while the remote
           task is push on the remote queue only. This task is never shared between multiple queue !
        */
        if (task->body != kaapi_tasksteal_body)
        {
          argsteal = (kaapi_tasksteal_arg_t*)kaapi_thread_pushdata(thread, sizeof(kaapi_tasksteal_arg_t));
          argsteal->origin_task           = task;
          argsteal->origin_body           = task->body;
          argsteal->origin_fmt            = 0;
          argsteal->war_param             = 0;    /* assume no war param: must be computed by thief ? */
          argsteal->cw_param              = 0;    /* assume no cw param: must be computed by thief ? */
          
          tasksteal = kaapi_thread_toptask(thread);
          kaapi_task_init( 
            tasksteal,
            kaapi_tasksteal_body,
            argsteal
          );
          ((kaapi_task_t* volatile)task)->reserved = tasksteal;
        }
        else {
          tasksteal = kaapi_thread_toptask(thread);
          kaapi_task_init( 
            tasksteal,
            task->body,
            task->sp
          );
          ((kaapi_task_t* volatile)task)->reserved = tasksteal;
        }

        kaapi_thread_pushtask(thread);
        return 0;
      }
    }
  } // for
  return EAGAIN;
}

